# mypy: disable-error-code="overload-overlap,misc,type-arg"

import datetime as dt
import sys

from collections.abc import Callable
from decimal import Decimal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
from typing import (
    Any,
    Collection,
    Generic,
    Iterable,
    Iterator,
    Literal,
    TypeVar,
    overload,
)

import numpy as np
import pandas as pd

from pandas.core.dtypes.base import ExtensionDtype
from pyarrow._compute import CastOptions
from pyarrow._stubs_typing import (
    ArrayLike,
    Indices,
    Mask,
    Order,
    SupportArrowArray,
    SupportArrowDeviceArray,
)
from pyarrow.lib import Buffer, MemoryPool, MonthDayNano, Tensor, _Weakrefable

from . import scalar, types
from .device import DeviceAllocationType
from .scalar import Scalar
from .types import (
    DataType,
    Field,
    MapType,
    _AsPyType,
    _BasicDataType,
    _BasicValueT,
    _DataType_CoT,
    _DataTypeT,
    _IndexT,
    _RunEndType,
    _Size,
    _ValueT,
)

_T = TypeVar("_T")

NullableIterable: TypeAlias = Iterable[_T | None]

@overload  # type: ignore[overload-overlap]
def array(
    values: NullableIterable[bool],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> BooleanArray: ...
@overload
def array(
    values: NullableIterable[int],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Int64Array: ...
@overload
def array(
    values: NullableIterable[float],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> DoubleArray: ...
@overload
def array(
    values: NullableIterable[Decimal],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Decimal128Array: ...
@overload
def array(
    values: NullableIterable[dict[str, Any]],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> StructArray: ...
@overload
def array(
    values: NullableIterable[dt.date],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Date32Array: ...
@overload
def array(
    values: NullableIterable[dt.time],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Time64Array: ...
@overload
def array(
    values: NullableIterable[dt.timedelta],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> DurationArray: ...
@overload
def array(
    values: NullableIterable[MonthDayNano],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> MonthDayNanoIntervalArray: ...
@overload
def array(
    values: NullableIterable[str],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> StringArray: ...
@overload
def array(
    values: NullableIterable[bytes],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> BinaryArray: ...
@overload
def array(
    values: NullableIterable[list],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> ListArray: ...
@overload
def array(
    values: NullableIterable[_Scalar_CoT],
    type: None = None,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Array[_Scalar_CoT]: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["null"] | types.NullType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> NullArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["bool", "boolean"] | types.BoolType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> BooleanArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["i1", "int8"] | types.Int8Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Int8Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["i2", "int16"] | types.Int16Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Int16Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["i4", "int32"] | types.Int32Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Int32Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["i8", "int64"] | types.Int64Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Int64Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["u1", "uint8"] | types.Int8Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> UInt8Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["u2", "uint16"] | types.Uint16Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> UInt16Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["u4", "uint32"] | types.Uint32Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> UInt32Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["u8", "uint64"] | types.Uint64Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> UInt64Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["f2", "halffloat", "float16"] | types.Float16Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> HalfFloatArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["f4", "float", "float32"] | types.Float32Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> FloatArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["f8", "double", "float64"] | types.Float64Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> DoubleArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["string", "str", "utf8"] | types.StringType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> StringArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["binary"] | types.BinaryType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> BinaryArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["large_string", "large_str", "large_utf8"] | types.LargeStringType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> LargeStringArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["large_binary"] | types.LargeBinaryType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> LargeBinaryArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["binary_view"] | types.BinaryViewType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> BinaryViewArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["string_view"] | types.StringViewType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> StringViewArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["date32", "date32[day]"] | types.Date32Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Date32Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["date64", "date64[ms]"] | types.Date64Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Date64Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["time32[s]", "time32[ms]"] | types.Time32Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Time32Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["time64[us]", "time64[ns]"] | types.Time64Type,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Time64Array: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["timestamp[s]", "timestamp[ms]", "timestamp[us]"] | types.TimestampType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> TimestampArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["duration[s]", "duration[ms]", "duration[us]", "duration[ns]"]
    | types.DurationType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> DurationArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["month_day_nano_interval"] | types.MonthDayNanoIntervalType,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> MonthDayNanoIntervalArray: ...
@overload
def array(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: _DataTypeT,
    mask: Mask | None = None,
    size: int | None = None,
    from_pandas: bool | None = None,
    safe: bool = True,
    memory_pool: MemoryPool | None = None,
) -> Array[Scalar[_DataTypeT]]: ...
@overload
def asarray(values: NullableIterable[bool]) -> BooleanArray: ...
@overload
def asarray(values: NullableIterable[int]) -> Int64Array: ...
@overload
def asarray(values: NullableIterable[float]) -> DoubleArray: ...
@overload
def asarray(values: NullableIterable[Decimal]) -> Decimal128Array: ...
@overload
def asarray(values: NullableIterable[dict[str, Any]]) -> StructArray: ...
@overload
def asarray(values: NullableIterable[dt.date]) -> Date32Array: ...
@overload
def asarray(values: NullableIterable[dt.time]) -> Time64Array: ...
@overload
def asarray(values: NullableIterable[dt.timedelta]) -> DurationArray: ...
@overload
def asarray(values: NullableIterable[MonthDayNano]) -> MonthDayNanoIntervalArray: ...
@overload
def asarray(values: NullableIterable[list]) -> ListArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["null"] | types.NullType,
) -> NullArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["bool", "boolean"] | types.BoolType,
) -> BooleanArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["i1", "int8"] | types.Int8Type,
) -> Int8Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["i2", "int16"] | types.Int16Type,
) -> Int16Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["i4", "int32"] | types.Int32Type,
) -> Int32Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["i8", "int64"] | types.Int64Type,
) -> Int64Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["u1", "uint8"] | types.Uint8Type,
) -> UInt8Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["u2", "uint16"] | types.Uint16Type,
) -> UInt16Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["u4", "uint32"] | types.Uint32Type,
) -> UInt32Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["u8", "uint64"] | types.Uint64Type,
) -> UInt64Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["f2", "halffloat", "float16"] | types.Float16Type,
) -> HalfFloatArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["f4", "float", "float32"] | types.Float32Type,
) -> FloatArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["f8", "double", "float64"] | types.Float64Type,
) -> DoubleArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["string", "str", "utf8"] | types.StringType,
) -> StringArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["binary"] | types.BinaryType,
) -> BinaryArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["large_string", "large_str", "large_utf8"] | types.LargeStringType,
) -> LargeStringArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["large_binary"] | types.LargeBinaryType,
) -> LargeBinaryArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["binary_view"] | types.BinaryViewType,
) -> BinaryViewArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["string_view"] | types.StringViewType,
) -> StringViewArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["date32", "date32[day]"] | types.Date32Type,
) -> Date32Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["date64", "date64[ms]"] | types.Date64Type,
) -> Date64Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["time32[s]", "time32[ms]"] | types.Time32Type,
) -> Time32Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["time64[us]", "time64[ns]"] | types.Time64Type,
) -> Time64Array: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["timestamp[s]", "timestamp[ms]", "timestamp[us]", "timestamp[ns]"]
    | types.TimestampType,
) -> TimestampArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["duration[s]", "duration[ms]", "duration[us]", "duration[ns]"]
    | types.DurationType,
) -> DurationArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: Literal["month_day_nano_interval"] | types.MonthDayNanoIntervalType,
) -> MonthDayNanoIntervalArray: ...
@overload
def asarray(
    values: Iterable | SupportArrowArray | SupportArrowDeviceArray,
    type: _DataTypeT,
) -> Array[Scalar[_DataTypeT]]: ...
@overload
def nulls(size: int, memory_pool: MemoryPool | None = None) -> NullArray: ...
@overload
def nulls(
    size: int, type: types.NullType | None, memory_pool: MemoryPool | None = None
) -> NullArray: ...
@overload
def nulls(
    size: int, type: types.BoolType, memory_pool: MemoryPool | None = None
) -> BooleanArray: ...
@overload
def nulls(size: int, type: types.Int8Type, memory_pool: MemoryPool | None = None) -> Int8Array: ...
@overload
def nulls(
    size: int, types: types.Int16Type, memory_pool: MemoryPool | None = None
) -> Int16Array: ...
@overload
def nulls(
    size: int, types: types.Int32Type, memory_pool: MemoryPool | None = None
) -> Int32Array: ...
@overload
def nulls(
    size: int, types: types.Int64Type, memory_pool: MemoryPool | None = None
) -> Int64Array: ...
@overload
def nulls(
    size: int, types: types.Uint8Type, memory_pool: MemoryPool | None = None
) -> UInt8Array: ...
@overload
def nulls(
    size: int, types: types.Uint16Type, memory_pool: MemoryPool | None = None
) -> UInt16Array: ...
@overload
def nulls(
    size: int, types: types.Uint32Type, memory_pool: MemoryPool | None = None
) -> UInt32Array: ...
@overload
def nulls(
    size: int, types: types.Uint64Type, memory_pool: MemoryPool | None = None
) -> UInt64Array: ...
@overload
def nulls(
    size: int, types: types.Float16Type, memory_pool: MemoryPool | None = None
) -> HalfFloatArray: ...
@overload
def nulls(
    size: int, types: types.Float32Type, memory_pool: MemoryPool | None = None
) -> FloatArray: ...
@overload
def nulls(
    size: int, types: types.Float64Type, memory_pool: MemoryPool | None = None
) -> DoubleArray: ...
@overload
def nulls(
    size: int, types: types.Decimal128Type, memory_pool: MemoryPool | None = None
) -> Decimal128Array: ...
@overload
def nulls(
    size: int, types: types.Decimal256Type, memory_pool: MemoryPool | None = None
) -> Decimal256Array: ...
@overload
def nulls(
    size: int, types: types.Date32Type, memory_pool: MemoryPool | None = None
) -> Date32Array: ...
@overload
def nulls(
    size: int, types: types.Date64Type, memory_pool: MemoryPool | None = None
) -> Date64Array: ...
@overload
def nulls(
    size: int, types: types.Time32Type, memory_pool: MemoryPool | None = None
) -> Time32Array: ...
@overload
def nulls(
    size: int, types: types.Time64Type, memory_pool: MemoryPool | None = None
) -> Time64Array: ...
@overload
def nulls(
    size: int, types: types.TimestampType, memory_pool: MemoryPool | None = None
) -> TimestampArray: ...
@overload
def nulls(
    size: int, types: types.DurationType, memory_pool: MemoryPool | None = None
) -> DurationArray: ...
@overload
def nulls(
    size: int, types: types.MonthDayNanoIntervalType, memory_pool: MemoryPool | None = None
) -> MonthDayNanoIntervalArray: ...
@overload
def nulls(
    size: int,
    types: types.BinaryType,
    memory_pool: MemoryPool | None = None,
) -> BinaryArray: ...
@overload
def nulls(
    size: int,
    types: types.LargeBinaryType,
    memory_pool: MemoryPool | None = None,
) -> LargeBinaryArray: ...
@overload
def nulls(
    size: int,
    types: types.FixedSizeBinaryType,
    memory_pool: MemoryPool | None = None,
) -> FixedSizeBinaryArray: ...
@overload
def nulls(
    size: int,
    types: types.StringType,
    memory_pool: MemoryPool | None = None,
) -> StringArray: ...
@overload
def nulls(
    size: int,
    types: types.LargeStringType,
    memory_pool: MemoryPool | None = None,
) -> LargeStringArray: ...
@overload
def nulls(
    size: int,
    types: types.BinaryViewType,
    memory_pool: MemoryPool | None = None,
) -> BinaryViewArray: ...
@overload
def nulls(
    size: int,
    types: types.StringViewType,
    memory_pool: MemoryPool | None = None,
) -> StringViewArray: ...
@overload
def nulls(
    size: int,
    types: types.ListType[_DataTypeT],
    memory_pool: MemoryPool | None = None,
) -> ListArray[scalar.ListScalar[_DataTypeT]]: ...
@overload
def nulls(
    size: int,
    types: types.FixedSizeListType[_DataTypeT, _Size],
    memory_pool: MemoryPool | None = None,
) -> FixedSizeListArray[_DataTypeT, _Size]: ...
@overload
def nulls(
    size: int,
    types: types.LargeListType[_DataTypeT],
    memory_pool: MemoryPool | None = None,
) -> LargeListArray[_DataTypeT]: ...
@overload
def nulls(
    size: int,
    types: types.ListViewType[_DataTypeT],
    memory_pool: MemoryPool | None = None,
) -> ListViewArray[_DataTypeT]: ...
@overload
def nulls(
    size: int,
    types: types.LargeListViewType[_DataTypeT],
    memory_pool: MemoryPool | None = None,
) -> LargeListViewArray[_DataTypeT]: ...
@overload
def nulls(
    size: int,
    types: types.StructType,
    memory_pool: MemoryPool | None = None,
) -> StructArray: ...
@overload
def nulls(
    size: int,
    types: types.MapType[_MapKeyT, _MapItemT],
    memory_pool: MemoryPool | None = None,
) -> MapArray[_MapKeyT, _MapItemT]: ...
@overload
def nulls(
    size: int,
    types: types.DictionaryType[_IndexT, _BasicValueT],
    memory_pool: MemoryPool | None = None,
) -> DictionaryArray[_IndexT, _BasicValueT]: ...
@overload
def nulls(
    size: int,
    types: types.RunEndEncodedType[_RunEndType, _BasicValueT],
    memory_pool: MemoryPool | None = None,
) -> RunEndEncodedArray[_RunEndType, _BasicValueT]: ...
@overload
def nulls(
    size: int,
    types: types.UnionType,
    memory_pool: MemoryPool | None = None,
) -> UnionArray: ...
@overload
def nulls(
    size: int,
    types: types.FixedShapeTensorType,
    memory_pool: MemoryPool | None = None,
) -> FixedShapeTensorArray: ...
@overload
def nulls(
    size: int,
    types: types.ExtensionType,
    memory_pool: MemoryPool | None = None,
) -> ExtensionArray: ...
@overload
def repeat(
    value: None | scalar.NullScalar, size: int, memory_pool: MemoryPool | None = None
) -> NullArray: ...
@overload
def repeat(  # type: ignore[overload-overlap]
    value: bool | scalar.BooleanScalar, size: int, memory_pool: MemoryPool | None = None
) -> BooleanArray: ...
@overload
def repeat(
    value: scalar.Int8Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Int8Array: ...
@overload
def repeat(
    value: scalar.Int16Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Int16Array: ...
@overload
def repeat(
    value: scalar.Int32Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Int32Array: ...
@overload
def repeat(
    value: int | scalar.Int64Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Int64Array: ...
@overload
def repeat(
    value: scalar.UInt8Scalar, size: int, memory_pool: MemoryPool | None = None
) -> UInt8Array: ...
@overload
def repeat(
    value: scalar.UInt16Scalar, size: int, memory_pool: MemoryPool | None = None
) -> UInt16Array: ...
@overload
def repeat(
    value: scalar.UInt32Scalar, size: int, memory_pool: MemoryPool | None = None
) -> UInt32Array: ...
@overload
def repeat(
    value: scalar.UInt64Scalar, size: int, memory_pool: MemoryPool | None = None
) -> UInt64Array: ...
@overload
def repeat(
    value: scalar.HalfFloatScalar, size: int, memory_pool: MemoryPool | None = None
) -> HalfFloatArray: ...
@overload
def repeat(
    value: scalar.FloatScalar, size: int, memory_pool: MemoryPool | None = None
) -> FloatArray: ...
@overload
def repeat(
    value: float | scalar.DoubleScalar, size: int, memory_pool: MemoryPool | None = None
) -> DoubleArray: ...
@overload
def repeat(
    value: Decimal | scalar.Decimal128Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Decimal128Array: ...
@overload
def repeat(
    value: scalar.Decimal256Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Decimal256Array: ...
@overload
def repeat(
    value: dt.date | scalar.Date32Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Date32Array: ...
@overload
def repeat(
    value: scalar.Date64Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Date64Array: ...
@overload
def repeat(
    value: scalar.Time32Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Time32Array: ...
@overload
def repeat(
    value: dt.time | scalar.Time64Scalar, size: int, memory_pool: MemoryPool | None = None
) -> Time64Array: ...
@overload
def repeat(
    value: scalar.TimestampScalar, size: int, memory_pool: MemoryPool | None = None
) -> TimestampArray: ...
@overload
def repeat(
    value: dt.timedelta | scalar.DurationScalar, size: int, memory_pool: MemoryPool | None = None
) -> DurationArray: ...
@overload
def repeat(  # type: ignore[overload-overlap]
    value: MonthDayNano | scalar.MonthDayNanoIntervalScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> MonthDayNanoIntervalArray: ...
@overload
def repeat(
    value: bytes | scalar.BinaryScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> BinaryArray: ...
@overload
def repeat(
    value: scalar.LargeBinaryScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> LargeBinaryArray: ...
@overload
def repeat(
    value: scalar.FixedSizeBinaryScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> FixedSizeBinaryArray: ...
@overload
def repeat(
    value: str | scalar.StringScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> StringArray: ...
@overload
def repeat(
    value: scalar.LargeStringScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> LargeStringArray: ...
@overload
def repeat(
    value: scalar.BinaryViewScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> BinaryViewArray: ...
@overload
def repeat(
    value: scalar.StringViewScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> StringViewArray: ...
@overload
def repeat(
    value: list | tuple | scalar.ListScalar[_DataTypeT],
    size: int,
    memory_pool: MemoryPool | None = None,
) -> ListArray[scalar.ListScalar[_DataTypeT]]: ...
@overload
def repeat(
    value: scalar.FixedSizeListScalar[_DataTypeT, _Size],
    size: int,
    memory_pool: MemoryPool | None = None,
) -> FixedSizeListArray[_DataTypeT, _Size]: ...
@overload
def repeat(
    value: scalar.LargeListScalar[_DataTypeT],
    size: int,
    memory_pool: MemoryPool | None = None,
) -> LargeListArray[_DataTypeT]: ...
@overload
def repeat(
    value: scalar.ListViewScalar[_DataTypeT],
    size: int,
    memory_pool: MemoryPool | None = None,
) -> ListViewArray[_DataTypeT]: ...
@overload
def repeat(
    value: scalar.LargeListViewScalar[_DataTypeT],
    size: int,
    memory_pool: MemoryPool | None = None,
) -> LargeListViewArray[_DataTypeT]: ...
@overload
def repeat(
    value: dict[str, Any] | scalar.StructScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> StructArray: ...
@overload
def repeat(
    value: scalar.MapScalar[_MapKeyT, _MapItemT],
    size: int,
    memory_pool: MemoryPool | None = None,
) -> MapArray[_MapKeyT, _MapItemT]: ...
@overload
def repeat(
    value: scalar.DictionaryScalar[_IndexT, _BasicValueT],
    size: int,
    memory_pool: MemoryPool | None = None,
) -> DictionaryArray[_IndexT, _BasicValueT]: ...
@overload
def repeat(
    value: scalar.RunEndEncodedScalar[_RunEndType, _BasicValueT],
    size: int,
    memory_pool: MemoryPool | None = None,
) -> RunEndEncodedArray[_RunEndType, _BasicValueT]: ...
@overload
def repeat(
    value: scalar.UnionScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> UnionArray: ...
@overload
def repeat(
    value: scalar.FixedShapeTensorScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> FixedShapeTensorArray: ...
@overload
def repeat(
    value: scalar.ExtensionScalar,
    size: int,
    memory_pool: MemoryPool | None = None,
) -> ExtensionArray: ...
def infer_type(values: Iterable, mask: Mask, from_pandas: bool = False) -> DataType: ...

_ConvertAs = TypeVar("_ConvertAs", pd.DataFrame, pd.Series)

class _PandasConvertible(_Weakrefable, Generic[_ConvertAs]):
    def to_pandas(
        self,
        memory_pool: MemoryPool | None = None,
        categories: list | None = None,
        strings_to_categorical: bool = False,
        zero_copy_only: bool = False,
        integer_object_nulls: bool = False,
        date_as_object: bool = True,
        timestamp_as_object: bool = False,
        use_threads: bool = True,
        deduplicate_objects: bool = True,
        ignore_metadata: bool = False,
        safe: bool = True,
        split_blocks: bool = False,
        self_destruct: bool = False,
        maps_as_pydicts: Literal["None", "lossy", "strict"] | None = None,
        types_mapper: Callable[[DataType], ExtensionDtype | None] | None = None,
        coerce_temporal_nanoseconds: bool = False,
    ) -> _ConvertAs: ...

_CastAs = TypeVar("_CastAs", bound=DataType)
_ScalarT = TypeVar("_ScalarT", bound=Scalar)
_Scalar_CoT = TypeVar("_Scalar_CoT", bound=Scalar, covariant=True)

class Array(_PandasConvertible[pd.Series], Generic[_Scalar_CoT]):
    def diff(self, other: Self) -> str: ...
    def cast(
        self,
        target_type: _CastAs,
        safe: bool = True,
        options: CastOptions | None = None,
        memory_pool: MemoryPool | None = None,
    ) -> Array[Scalar[_CastAs]]: ...
    def view(self, target_type: _CastAs) -> Array[Scalar[_CastAs]]: ...
    def sum(self, **kwargs) -> _Scalar_CoT: ...
    @property
    def type(self: Array[Scalar[_DataType_CoT]]) -> _DataType_CoT: ...
    def unique(self) -> Self: ...
    def dictionary_encode(self, null_encoding: str = "mask") -> DictionaryArray: ...
    @overload
    @staticmethod
    def from_pandas(
        obj: pd.Series | np.ndarray | ArrayLike,
        *,
        mask: Mask | None = None,
        type: _DataTypeT,
        safe: bool = True,
        memory_pool: MemoryPool | None = None,
    ) -> Array[Scalar[_DataTypeT]]: ...
    @overload
    @staticmethod
    def from_pandas(
        obj: pd.Series | np.ndarray | ArrayLike,
        *,
        mask: Mask | None = None,
        safe: bool = True,
        memory_pool: MemoryPool | None = None,
    ) -> Array[Scalar]: ...
    @staticmethod
    def from_buffers(
        type: _DataTypeT,
        length: int,
        buffers: list[Buffer],
        null_count: int = -1,
        offset=0,
        children: Collection[Array[Scalar[_DataTypeT]]] | None = None,
    ) -> Array[Scalar[_DataTypeT]]: ...
    @property
    def null_count(self) -> int: ...
    @property
    def nbytes(self) -> int: ...
    def get_total_buffer_size(self) -> int: ...
    def __sizeof__(self) -> int: ...
    def __iter__(self) -> Iterator[_Scalar_CoT]: ...
    def to_string(
        self,
        *,
        indent: int = 2,
        top_level_indent: int = 0,
        window: int = 10,
        container_window: int = 2,
        skip_new_lines: bool = False,
    ) -> str: ...
    format = to_string
    def equals(self, other: Self) -> bool: ...
    def __len__(self) -> int: ...
    def is_null(self, *, nan_is_null: bool = False) -> BooleanArray: ...
    def is_nan(self) -> BooleanArray: ...
    def is_valid(self) -> BooleanArray: ...
    def fill_null(
        self: Array[Scalar[_BasicDataType[_AsPyType]]], fill_value: _AsPyType
    ) -> Array[Scalar[_BasicDataType[_AsPyType]]]: ...
    @overload
    def __getitem__(self, key: int) -> _Scalar_CoT: ...
    @overload
    def __getitem__(self, key: slice) -> Self: ...
    def slice(self, offset: int = 0, length: int | None = None) -> Self: ...
    def take(self, indices: Indices) -> Self: ...
    def drop_null(self) -> Self: ...
    def filter(
        self,
        mask: Mask,
        *,
        null_selection_behavior: Literal["drop", "emit_null"] = "drop",
    ) -> Self: ...
    @overload
    def index(
        self: Array[_ScalarT],
        value: _ScalarT,
        start: int | None = None,
        end: int | None = None,
        *,
        memory_pool: MemoryPool | None = None,
    ) -> scalar.Int64Scalar: ...
    @overload
    def index(
        self: Array[Scalar[_BasicDataType[_AsPyType]]],
        value: _AsPyType,
        start: int | None = None,
        end: int | None = None,
        *,
        memory_pool: MemoryPool | None = None,
    ) -> scalar.Int64Scalar: ...
    def sort(self, order: Order = "ascending", **kwargs) -> Self: ...
    def __array__(self, dtype: np.dtype | None = None, copy: bool | None = None) -> np.ndarray: ...
    def to_numpy(self, zero_copy_only: bool = True, writable: bool = False) -> np.ndarray: ...
    def to_pylist(
        self: Array[Scalar[_BasicDataType[_AsPyType]]],
    ) -> list[_AsPyType | None]: ...
    tolist = to_pylist
    def validate(self, *, full: bool = False) -> None: ...
    @property
    def offset(self) -> int: ...
    def buffers(self) -> list[Buffer | None]: ...
    def _export_to_c(self, out_ptr: int, out_schema_ptr: int = 0) -> None: ...
    @classmethod
    def _import_from_c(cls, in_ptr: int, type: int | DataType) -> Self: ...
    def __arrow_c_array__(self, requested_schema=None) -> Any: ...
    @classmethod
    def _import_from_c_capsule(cls, schema_capsule, array_capsule) -> Self: ...
    def _export_to_c_device(self, out_ptr: int, out_schema_ptr: int = 0) -> None: ...
    @classmethod
    def _import_from_c_device(cls, in_ptr: int, type: DataType | int) -> Self: ...
    def __arrow_c_device_array__(self, requested_schema=None, **kwargs) -> Any: ...
    @classmethod
    def _import_from_c_device_capsule(cls, schema_capsule, array_capsule) -> Self: ...
    def __dlpack__(self, stream: int | None = None) -> Any: ...
    def __dlpack_device__(self) -> tuple[int, int]: ...
    @property
    def device_type(self) -> DeviceAllocationType: ...
    @property
    def is_cpu(self) -> bool: ...

class NullArray(Array[scalar.NullScalar]): ...

class BooleanArray(Array[scalar.BooleanScalar]):
    @property
    def false_count(self) -> int: ...
    @property
    def true_count(self) -> int: ...

class NumericArray(Array[_Scalar_CoT]): ...
class IntegerArray(NumericArray[_Scalar_CoT]): ...
class FloatingPointArray(NumericArray[_Scalar_CoT]): ...
class Int8Array(IntegerArray[scalar.Int8Scalar]): ...
class UInt8Array(IntegerArray[scalar.UInt8Scalar]): ...
class Int16Array(IntegerArray[scalar.Int16Scalar]): ...
class UInt16Array(IntegerArray[scalar.UInt16Scalar]): ...
class Int32Array(IntegerArray[scalar.Int32Scalar]): ...
class UInt32Array(IntegerArray[scalar.UInt32Scalar]): ...
class Int64Array(IntegerArray[scalar.Int64Scalar]): ...
class UInt64Array(IntegerArray[scalar.UInt64Scalar]): ...
class Date32Array(NumericArray[scalar.Date32Scalar]): ...
class Date64Array(NumericArray[scalar.Date64Scalar]): ...
class TimestampArray(NumericArray[scalar.TimestampScalar]): ...
class Time32Array(NumericArray[scalar.Time32Scalar]): ...
class Time64Array(NumericArray[scalar.Time64Scalar]): ...
class DurationArray(NumericArray[scalar.DurationScalar]): ...
class MonthDayNanoIntervalArray(Array[scalar.MonthDayNanoIntervalScalar]): ...
class HalfFloatArray(FloatingPointArray[scalar.HalfFloatScalar]): ...
class FloatArray(FloatingPointArray[scalar.FloatScalar]): ...
class DoubleArray(FloatingPointArray[scalar.DoubleScalar]): ...
class FixedSizeBinaryArray(Array[scalar.FixedSizeBinaryScalar]): ...
class Decimal128Array(FixedSizeBinaryArray): ...
class Decimal256Array(FixedSizeBinaryArray): ...

class BaseListArray(Array[_Scalar_CoT]):
    def flatten(self, recursive: bool = False) -> Array: ...
    def value_parent_indices(self) -> Int64Array: ...
    def value_lengths(self) -> Int32Array: ...

class ListArray(BaseListArray[_Scalar_CoT]):
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int32Array,
        values: Array[Scalar[_DataTypeT]],
        *,
        type: None = None,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> ListArray[scalar.ListScalar[_DataTypeT]]: ...
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int32Array,
        values: Array,
        *,
        type: _DataTypeT,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> ListArray[scalar.ListScalar[_DataTypeT]]: ...
    @property
    def values(self) -> Array: ...
    @property
    def offsets(self) -> Int32Array: ...

class LargeListArray(BaseListArray[scalar.LargeListScalar[_DataType_CoT]]):
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int64Array,
        values: Array[Scalar[_DataTypeT]],
        *,
        type: None = None,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> LargeListArray[_DataTypeT]: ...
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int64Array,
        values: Array,
        *,
        type: _DataTypeT,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> LargeListArray[_DataTypeT]: ...
    @property
    def values(self) -> Array: ...
    @property
    def offsets(self) -> Int64Array: ...

class ListViewArray(BaseListArray[scalar.ListViewScalar[_DataType_CoT]]):
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int32Array,
        values: Array[Scalar[_DataTypeT]],
        *,
        type: None = None,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> ListViewArray[_DataTypeT]: ...
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int32Array,
        values: Array,
        *,
        type: _DataTypeT,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> ListViewArray[_DataTypeT]: ...
    @property
    def values(self) -> Array: ...
    @property
    def offsets(self) -> Int32Array: ...
    @property
    def sizes(self) -> Int32Array: ...

class LargeListViewArray(BaseListArray[scalar.LargeListScalar[_DataType_CoT]]):
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int64Array,
        values: Array[Scalar[_DataTypeT]],
        *,
        type: None = None,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> LargeListViewArray[_DataTypeT]: ...
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int64Array,
        values: Array,
        *,
        type: _DataTypeT,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> LargeListViewArray[_DataTypeT]: ...
    @property
    def values(self) -> Array: ...
    @property
    def offsets(self) -> Int64Array: ...
    @property
    def sizes(self) -> Int64Array: ...

class FixedSizeListArray(BaseListArray[scalar.FixedSizeListScalar[_DataType_CoT, _Size]]):
    @overload
    @classmethod
    def from_arrays(
        cls,
        values: Array[Scalar[_DataTypeT]],
        *,
        type: None = None,
        mask: Mask | None = None,
    ) -> FixedSizeListArray[_DataTypeT, None]: ...
    @overload
    @classmethod
    def from_arrays(
        cls,
        values: Array[Scalar[_DataTypeT]],
        limit_size: _Size,
        *,
        type: None = None,
        mask: Mask | None = None,
    ) -> FixedSizeListArray[_DataTypeT, _Size]: ...
    @property
    def values(self) -> BaseListArray[scalar.ListScalar[_DataType_CoT]]: ...

_MapKeyT = TypeVar("_MapKeyT", bound=_BasicDataType)
_MapItemT = TypeVar("_MapItemT", bound=_BasicDataType)

class MapArray(ListArray[scalar.MapScalar[_MapKeyT, _MapItemT]]):
    @overload  # type: ignore[override]
    @classmethod
    def from_arrays(
        cls,
        offsets: Int64Array,
        keys: Array[Scalar[_MapKeyT]],
        items: Array[Scalar[_MapItemT]],
        *,
        type: None = None,
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> MapArray[_MapKeyT, _MapItemT]: ...
    @overload
    @classmethod
    def from_arrays(
        cls,
        offsets: Int64Array,
        values: Array,
        *,
        type: MapType[_MapKeyT, _MapItemT],
        pool: MemoryPool | None = None,
        mask: Mask | None = None,
    ) -> MapArray[_MapKeyT, _MapItemT]: ...
    @property
    def keys(self) -> Array: ...
    @property
    def items(self) -> Array: ...

class UnionArray(Array[scalar.UnionScalar]):
    def child(self, pos: int) -> Field: ...
    def field(self, pos: int) -> Array: ...
    @property
    def type_codes(self) -> Int8Array: ...
    @property
    def offsets(self) -> Int32Array: ...
    @staticmethod
    def from_dense(
        types: Int8Array,
        value_offsets: Int32Array,
        children: Collection[Array],
        field_names: list[str] | None = None,
        type_codes: Int8Array | None = None,
    ) -> UnionArray: ...
    @staticmethod
    def from_sparse(
        types: Int8Array,
        children: Collection[Array],
        field_names: list[str] | None = None,
        type_codes: Int8Array | None = None,
    ) -> UnionArray: ...

class StringArray(Array[scalar.StringScalar]):
    @staticmethod
    def from_buffers(  # type: ignore[override]
        length: int,
        value_offsets: Buffer,
        data: Buffer,
        null_bitmap: Buffer | None = None,
        null_count: int | None = -1,
        offset: int | None = 0,
    ) -> StringArray: ...

class LargeStringArray(Array[scalar.LargeStringScalar]):
    @staticmethod
    def from_buffers(  # type: ignore[override]
        length: int,
        value_offsets: Buffer,
        data: Buffer,
        null_bitmap: Buffer | None = None,
        null_count: int | None = -1,
        offset: int | None = 0,
    ) -> StringArray: ...

class StringViewArray(Array[scalar.StringViewScalar]): ...

class BinaryArray(Array[scalar.BinaryScalar]):
    @property
    def total_values_length(self) -> int: ...

class LargeBinaryArray(Array[scalar.LargeBinaryScalar]):
    @property
    def total_values_length(self) -> int: ...

class BinaryViewArray(Array[scalar.BinaryViewScalar]): ...

class DictionaryArray(Array[scalar.DictionaryScalar[_IndexT, _BasicValueT]]):
    @staticmethod
    def from_buffers(  # type: ignore[override]
        type: _BasicValueT,
        length: int,
        buffers: list[Buffer],
        dictionary: Array | np.ndarray | pd.Series,
        null_count: int = -1,
        offset: int = 0,
    ) -> DictionaryArray[Any, _BasicValueT]: ...
    @staticmethod
    def from_arrays(
        indices: Indices,
        dictionary: Array | np.ndarray | pd.Series,
        mask: np.ndarray | pd.Series | BooleanArray | None = None,
        ordered: bool = False,
        from_pandas: bool = False,
        safe: bool = True,
        memory_pool: MemoryPool | None = None,
    ) -> DictionaryArray: ...

class StructArray(Array[scalar.StructScalar]):
    def field(self, index: int | str) -> Array: ...
    def flatten(self, memory_pool: MemoryPool | None = None) -> list[Array]: ...
    @staticmethod
    def from_arrays(
        arrays: Iterable[Array],
        names: list[str] | None = None,
        fields: list[Field] | None = None,
        mask=None,
        memory_pool: MemoryPool | None = None,
    ) -> StructArray: ...
    def sort(self, order: Order = "ascending", by: str | None = None, **kwargs) -> StructArray: ...

class RunEndEncodedArray(Array[scalar.RunEndEncodedScalar[_RunEndType, _BasicValueT]]):
    @overload
    @staticmethod
    def from_arrays(
        run_ends: Int16Array,
        values: Array,
        type: _ValueT | None = None,
    ) -> RunEndEncodedArray[types.Int16Type, _BasicValueT]: ...
    @overload
    @staticmethod
    def from_arrays(
        run_ends: Int32Array,
        values: Array,
        type: _ValueT | None = None,
    ) -> RunEndEncodedArray[types.Int32Type, _BasicValueT]: ...
    @overload
    @staticmethod
    def from_arrays(
        run_ends: Int64Array,
        values: Array,
        type: _ValueT | None = None,
    ) -> RunEndEncodedArray[types.Int64Type, _BasicValueT]: ...
    @staticmethod
    def from_buffers(  # type: ignore[override]
        type: _ValueT,
        length: int,
        buffers: list[Buffer],
        null_count: int = -1,
        offset=0,
        children: tuple[Array, Array] | None = None,
    ) -> RunEndEncodedArray[Any, _BasicValueT]: ...
    @property
    def run_ends(self) -> Array[scalar.Scalar[_RunEndType]]: ...
    @property
    def values(self) -> Array[scalar.Scalar[_BasicValueT]]: ...
    def find_physical_offset(self) -> int: ...
    def find_physical_length(self) -> int: ...

_ArrayT = TypeVar("_ArrayT", bound=Array)

class ExtensionArray(Array[scalar.ExtensionScalar], Generic[_ArrayT]):
    @property
    def storage(self) -> Any: ...
    @staticmethod
    def from_storage(
        typ: types.BaseExtensionType, storage: _ArrayT
    ) -> ExtensionArray[_ArrayT]: ...

class FixedShapeTensorArray(ExtensionArray[_ArrayT]):
    def to_numpy_ndarray(self) -> np.ndarray: ...
    def to_tensor(self) -> Tensor: ...
    @staticmethod
    def from_numpy_ndarray(obj: np.ndarray) -> FixedShapeTensorArray: ...

def concat_arrays(arrays: Iterable[_ArrayT], memory_pool: MemoryPool | None = None) -> _ArrayT: ...
def _empty_array(type: _DataTypeT) -> Array[scalar.Scalar[_DataTypeT]]: ...

__all__ = [
    "array",
    "asarray",
    "nulls",
    "repeat",
    "infer_type",
    "_PandasConvertible",
    "Array",
    "NullArray",
    "BooleanArray",
    "NumericArray",
    "IntegerArray",
    "FloatingPointArray",
    "Int8Array",
    "UInt8Array",
    "Int16Array",
    "UInt16Array",
    "Int32Array",
    "UInt32Array",
    "Int64Array",
    "UInt64Array",
    "Date32Array",
    "Date64Array",
    "TimestampArray",
    "Time32Array",
    "Time64Array",
    "DurationArray",
    "MonthDayNanoIntervalArray",
    "HalfFloatArray",
    "FloatArray",
    "DoubleArray",
    "FixedSizeBinaryArray",
    "Decimal128Array",
    "Decimal256Array",
    "BaseListArray",
    "ListArray",
    "LargeListArray",
    "ListViewArray",
    "LargeListViewArray",
    "FixedSizeListArray",
    "MapArray",
    "UnionArray",
    "StringArray",
    "LargeStringArray",
    "StringViewArray",
    "BinaryArray",
    "LargeBinaryArray",
    "BinaryViewArray",
    "DictionaryArray",
    "StructArray",
    "RunEndEncodedArray",
    "ExtensionArray",
    "FixedShapeTensorArray",
    "concat_arrays",
    "_empty_array",
]



from _typeshed import StrPath

from ._fs import FileSystem

class HadoopFileSystem(FileSystem):
    def __init__(
        self,
        host: str,
        port: int = 8020,
        *,
        user: str | None = None,
        replication: int = 3,
        buffer_size: int = 0,
        default_block_size: int | None = None,
        kerb_ticket: StrPath | None = None,
        extra_conf: dict | None = None,
    ): ...
    @staticmethod
    def from_uri(uri: str) -> HadoopFileSystem: ...  # type: ignore[override]



from pyarrow._flight import (
    Action,
    ActionType,
    BasicAuth,
    CallInfo,
    CertKeyPair,
    ClientAuthHandler,
    ClientMiddleware,
    ClientMiddlewareFactory,
    DescriptorType,
    FlightCallOptions,
    FlightCancelledError,
    FlightClient,
    FlightDataStream,
    FlightDescriptor,
    FlightEndpoint,
    FlightError,
    FlightInfo,
    FlightInternalError,
    FlightMetadataReader,
    FlightMetadataWriter,
    FlightMethod,
    FlightServerBase,
    FlightServerError,
    FlightStreamChunk,
    FlightStreamReader,
    FlightStreamWriter,
    FlightTimedOutError,
    FlightUnauthenticatedError,
    FlightUnauthorizedError,
    FlightUnavailableError,
    FlightWriteSizeExceededError,
    GeneratorStream,
    Location,
    MetadataRecordBatchReader,
    MetadataRecordBatchWriter,
    RecordBatchStream,
    Result,
    SchemaResult,
    ServerAuthHandler,
    ServerCallContext,
    ServerMiddleware,
    ServerMiddlewareFactory,
    Ticket,
    TracingServerMiddlewareFactory,
    connect,
)

__all__ = [
    "Action",
    "ActionType",
    "BasicAuth",
    "CallInfo",
    "CertKeyPair",
    "ClientAuthHandler",
    "ClientMiddleware",
    "ClientMiddlewareFactory",
    "DescriptorType",
    "FlightCallOptions",
    "FlightCancelledError",
    "FlightClient",
    "FlightDataStream",
    "FlightDescriptor",
    "FlightEndpoint",
    "FlightError",
    "FlightInfo",
    "FlightInternalError",
    "FlightMetadataReader",
    "FlightMetadataWriter",
    "FlightMethod",
    "FlightServerBase",
    "FlightServerError",
    "FlightStreamChunk",
    "FlightStreamReader",
    "FlightStreamWriter",
    "FlightTimedOutError",
    "FlightUnauthenticatedError",
    "FlightUnauthorizedError",
    "FlightUnavailableError",
    "FlightWriteSizeExceededError",
    "GeneratorStream",
    "Location",
    "MetadataRecordBatchReader",
    "MetadataRecordBatchWriter",
    "RecordBatchStream",
    "Result",
    "SchemaResult",
    "ServerAuthHandler",
    "ServerCallContext",
    "ServerMiddleware",
    "ServerMiddlewareFactory",
    "Ticket",
    "TracingServerMiddlewareFactory",
    "connect",
]



import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Any, Iterable, Sequence

from pyarrow.interchange.column import _PyArrowColumn
from pyarrow.lib import RecordBatch, Table

class _PyArrowDataFrame:
    def __init__(
        self, df: Table | RecordBatch, nan_as_null: bool = False, allow_copy: bool = True
    ) -> None: ...
    def __dataframe__(
        self, nan_as_null: bool = False, allow_copy: bool = True
    ) -> _PyArrowDataFrame: ...
    @property
    def metadata(self) -> dict[str, Any]: ...
    def num_columns(self) -> int: ...
    def num_rows(self) -> int: ...
    def num_chunks(self) -> int: ...
    def column_names(self) -> Iterable[str]: ...
    def get_column(self, i: int) -> _PyArrowColumn: ...
    def get_column_by_name(self, name: str) -> _PyArrowColumn: ...
    def get_columns(self) -> Iterable[_PyArrowColumn]: ...
    def select_columns(self, indices: Sequence[int]) -> Self: ...
    def select_columns_by_name(self, names: Sequence[str]) -> Self: ...
    def get_chunks(self, n_chunks: int | None = None) -> Iterable[Self]: ...



from io import IOBase

import pandas as pd
import pyarrow.lib as lib

from pyarrow.lib import (
    IpcReadOptions,
    IpcWriteOptions,
    Message,
    MessageReader,
    MetadataVersion,
    ReadStats,
    RecordBatchReader,
    WriteStats,
    _ReadPandasMixin,
    get_record_batch_size,
    get_tensor_size,
    read_message,
    read_record_batch,
    read_schema,
    read_tensor,
    write_tensor,
)

class RecordBatchStreamReader(lib._RecordBatchStreamReader):
    def __init__(
        self,
        source: bytes | lib.Buffer | lib.NativeFile | IOBase,
        *,
        options: IpcReadOptions | None = None,
        memory_pool: lib.MemoryPool | None = None,
    ) -> None: ...

class RecordBatchStreamWriter(lib._RecordBatchStreamWriter):
    def __init__(
        self,
        sink: str | lib.NativeFile | IOBase,
        schema: lib.Schema,
        *,
        use_legacy_format: bool | None = None,
        options: IpcWriteOptions | None = None,
    ) -> None: ...

class RecordBatchFileReader(lib._RecordBatchFileReader):
    def __init__(
        self,
        source: bytes | lib.Buffer | lib.NativeFile | IOBase,
        footer_offset: int | None = None,
        *,
        options: IpcReadOptions | None,
        memory_pool: lib.MemoryPool | None = None,
    ) -> None: ...

class RecordBatchFileWriter(lib._RecordBatchFileWriter):
    def __init__(
        self,
        sink: str | lib.NativeFile | IOBase,
        schema: lib.Schema,
        *,
        use_legacy_format: bool | None = None,
        options: IpcWriteOptions | None = None,
    ) -> None: ...

def new_stream(
    sink: str | lib.NativeFile | IOBase,
    schema: lib.Schema,
    *,
    use_legacy_format: bool | None = None,
    options: IpcWriteOptions | None = None,
) -> RecordBatchStreamWriter: ...
def open_stream(
    source: bytes | lib.Buffer | lib.NativeFile | IOBase,
    *,
    options: IpcReadOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> RecordBatchStreamReader: ...
def new_file(
    sink: str | lib.NativeFile | IOBase,
    schema: lib.Schema,
    *,
    use_legacy_format: bool | None = None,
    options: IpcWriteOptions | None = None,
) -> RecordBatchFileWriter: ...
def open_file(
    source: bytes | lib.Buffer | lib.NativeFile | IOBase,
    footer_offset: int | None = None,
    *,
    options: IpcReadOptions | None = None,
    memory_pool: lib.MemoryPool | None = None,
) -> RecordBatchFileReader: ...
def serialize_pandas(
    df: pd.DataFrame, *, nthreads: int | None = None, preserve_index: bool | None = None
) -> lib.Buffer: ...
def deserialize_pandas(buf: lib.Buffer, *, use_threads: bool = True) -> pd.DataFrame: ...

__all__ = [
    "IpcReadOptions",
    "IpcWriteOptions",
    "Message",
    "MessageReader",
    "MetadataVersion",
    "ReadStats",
    "RecordBatchReader",
    "WriteStats",
    "_ReadPandasMixin",
    "get_record_batch_size",
    "get_tensor_size",
    "read_message",
    "read_record_batch",
    "read_schema",
    "read_tensor",
    "write_tensor",
    "RecordBatchStreamReader",
    "RecordBatchStreamWriter",
    "RecordBatchFileReader",
    "RecordBatchFileWriter",
    "new_stream",
    "open_stream",
    "new_file",
    "open_file",
    "serialize_pandas",
    "deserialize_pandas",
]



from dataclasses import dataclass
from typing import IO, Any, Iterable, TypedDict

from _typeshed import StrPath

from ._compute import Expression
from ._dataset import (
    DatasetFactory,
    FileFormat,
    FileFragment,
    FileWriteOptions,
    Fragment,
    FragmentScanOptions,
    Partitioning,
    PartitioningFactory,
)
from ._dataset_parquet_encryption import ParquetDecryptionConfig
from ._fs import SupportedFileSystem
from ._parquet import FileDecryptionProperties, FileMetaData
from .lib import CacheOptions, Schema, _Weakrefable

parquet_encryption_enabled: bool

class ParquetFileFormat(FileFormat):
    def __init__(
        self,
        read_options: ParquetReadOptions,
        default_fragment_scan_options: ParquetFragmentScanOptions,
        **kwargs,
    ) -> None: ...
    @property
    def read_options(self) -> ParquetReadOptions: ...
    def make_write_options(self) -> ParquetFileWriteOptions: ...  # type: ignore[override]
    def equals(self, other: ParquetFileFormat) -> bool: ...
    @property
    def default_extname(self) -> str: ...
    def make_fragment(
        self,
        file: StrPath | IO,
        filesystem: SupportedFileSystem | None = None,
        partition_expression: Expression | None = None,
        row_groups: Iterable[int] | None = None,
        *,
        file_size: int | None = None,
    ) -> Fragment: ...

class _NameStats(TypedDict):
    min: Any
    max: Any

class RowGroupInfo:
    id: int
    metadata: FileMetaData
    schema: Schema

    def __init__(self, id: int, metadata: FileMetaData, schema: Schema) -> None: ...
    @property
    def num_rows(self) -> int: ...
    @property
    def total_byte_size(self) -> int: ...
    @property
    def statistics(self) -> dict[str, _NameStats]: ...

class ParquetFileFragment(FileFragment):
    def ensure_complete_metadata(self) -> None: ...
    @property
    def row_groups(self) -> list[RowGroupInfo]: ...
    @property
    def metadata(self) -> FileMetaData: ...
    @property
    def num_row_groups(self) -> int: ...
    def split_by_row_group(
        self, filter: Expression | None = None, schema: Schema | None = None
    ) -> list[Fragment]: ...
    def subset(
        self,
        filter: Expression | None = None,
        schema: Schema | None = None,
        row_group_ids: list[int] | None = None,
    ) -> ParquetFileFormat: ...

class ParquetReadOptions(_Weakrefable):
    def __init__(
        self, dictionary_columns: list[str] | None, coerce_int96_timestamp_unit: str | None = None
    ) -> None: ...
    @property
    def coerce_int96_timestamp_unit(self) -> str: ...
    @coerce_int96_timestamp_unit.setter
    def coerce_int96_timestamp_unit(self, unit: str) -> None: ...
    def equals(self, other: ParquetReadOptions) -> bool: ...

class ParquetFileWriteOptions(FileWriteOptions):
    def update(self, **kwargs) -> None: ...
    def _set_properties(self) -> None: ...
    def _set_arrow_properties(self) -> None: ...
    def _set_encryption_config(self) -> None: ...

@dataclass(kw_only=True)
class ParquetFragmentScanOptions(FragmentScanOptions):
    use_buffered_stream: bool = False
    buffer_size: int = 8192
    pre_buffer: bool = True
    cache_options: CacheOptions | None = None
    thrift_string_size_limit: int | None = None
    thrift_container_size_limit: int | None = None
    decryption_config: ParquetDecryptionConfig | None = None
    decryption_properties: FileDecryptionProperties | None = None
    page_checksum_verification: bool = False

    def equals(self, other: ParquetFragmentScanOptions) -> bool: ...

@dataclass
class ParquetFactoryOptions(_Weakrefable):
    partition_base_dir: str | None = None
    partitioning: Partitioning | PartitioningFactory | None = None
    validate_column_chunk_paths: bool = False

class ParquetDatasetFactory(DatasetFactory):
    def __init__(
        self,
        metadata_path: str,
        filesystem: SupportedFileSystem,
        format: FileFormat,
        options: ParquetFactoryOptions | None = None,
    ) -> None: ...



from collections.abc import Callable
from os import PathLike
from typing import Any, Protocol, Sequence, TypeVar

_F = TypeVar("_F", bound=Callable)
_N = TypeVar("_N")

class _DocStringComponents(Protocol):
    _docstring_components: list[str]

def doc(
    *docstrings: str | _DocStringComponents | Callable | None, **params: Any
) -> Callable[[_F], _F]: ...
def _is_iterable(obj) -> bool: ...
def _is_path_like(path) -> bool: ...
def _stringify_path(path: str | PathLike) -> str: ...
def product(seq: Sequence[_N]) -> _N: ...
def get_contiguous_span(
    shape: tuple[int, ...], strides: tuple[int, ...], itemsize: int
) -> tuple[int, int]: ...
def find_free_port() -> int: ...
def guid() -> str: ...
def download_tzdata_on_windows() -> None: ...
def _deprecate_api(old_name, new_name, api, next_version, type=...): ...
def _deprecate_class(old_name, new_class, next_version, instancecheck=True): ...



# mypy: disable-error-code="overload-overlap,misc,type-arg"
import collections.abc
import datetime as dt
import sys

from decimal import Decimal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
from typing import Any, Generic, Iterator, Mapping, overload

import numpy as np

from pyarrow._compute import CastOptions
from pyarrow.lib import Array, Buffer, MemoryPool, MonthDayNano, Tensor, _Weakrefable
from typing_extensions import TypeVar

from . import types
from .types import (
    _AsPyType,
    _DataType_CoT,
    _DataTypeT,
    _Time32Unit,
    _Time64Unit,
    _Tz,
    _Unit,
)

_AsPyTypeK = TypeVar("_AsPyTypeK")
_AsPyTypeV = TypeVar("_AsPyTypeV")

class Scalar(_Weakrefable, Generic[_DataType_CoT]):
    @property
    def type(self) -> _DataType_CoT: ...
    @property
    def is_valid(self) -> bool: ...
    @overload
    def cast(
        self,
        target_type: None,
        safe: bool = True,
        options: CastOptions | None = None,
        memory_pool: MemoryPool | None = None,
    ) -> Self: ...
    @overload
    def cast(
        self,
        target_type: _DataTypeT,
        safe: bool = True,
        options: CastOptions | None = None,
        memory_pool: MemoryPool | None = None,
    ) -> Scalar[_DataTypeT]: ...
    def validate(self, *, full: bool = False) -> None: ...
    def equals(self, other: Scalar) -> bool: ...
    def __hash__(self) -> int: ...
    @overload
    def as_py(self: Scalar[types._BasicDataType[_AsPyType]]) -> _AsPyType: ...
    @overload
    def as_py(
        self: Scalar[types.ListType[types._BasicDataType[_AsPyType]]],
    ) -> list[_AsPyType]: ...
    @overload
    def as_py(
        self: Scalar[
            types.ListType[
                types.DictionaryType[types._IndexT, types._BasicDataType[_AsPyTypeV], Any]
            ]
        ],
    ) -> list[dict[_AsPyTypeK, _AsPyTypeV]]: ...
    @overload
    def as_py(
        self: Scalar[
            types.ListType[types.DictionaryType[Any, types._BasicDataType[_AsPyTypeV], Any]],
        ],
    ) -> list[dict[Any, _AsPyTypeV]]: ...
    @overload
    def as_py(
        self: Scalar[types.ListType[types.DictionaryType[types._IndexT, Any, Any]],],
    ) -> list[dict[_AsPyTypeK, Any]]: ...
    @overload
    def as_py(
        self: Scalar[types.StructType],
    ) -> list[dict[str, Any]]: ...
    @overload
    def as_py(
        self: Scalar[
            types.MapType[types._BasicDataType[_AsPyTypeK], types._BasicDataType[_AsPyTypeV]]
        ],
    ) -> list[tuple[_AsPyTypeK, _AsPyTypeV]]: ...
    @overload
    def as_py(
        self: Scalar[types.MapType[Any, types._BasicDataType[_AsPyTypeV]]],
    ) -> list[tuple[Any, _AsPyTypeV]]: ...
    @overload
    def as_py(
        self: Scalar[types.MapType[types._BasicDataType[_AsPyTypeK], Any]],
    ) -> list[tuple[_AsPyTypeK, Any]]: ...
    @overload
    def as_py(self: Scalar[Any]) -> Any: ...

_NULL: TypeAlias = None
NA = _NULL

class NullScalar(Scalar[types.NullType]): ...
class BooleanScalar(Scalar[types.BoolType]): ...
class UInt8Scalar(Scalar[types.Uint8Type]): ...
class Int8Scalar(Scalar[types.Int8Type]): ...
class UInt16Scalar(Scalar[types.Uint16Type]): ...
class Int16Scalar(Scalar[types.Int16Type]): ...
class UInt32Scalar(Scalar[types.Uint32Type]): ...
class Int32Scalar(Scalar[types.Int32Type]): ...
class UInt64Scalar(Scalar[types.Uint64Type]): ...
class Int64Scalar(Scalar[types.Int64Type]): ...
class HalfFloatScalar(Scalar[types.Float16Type]): ...
class FloatScalar(Scalar[types.Float32Type]): ...
class DoubleScalar(Scalar[types.Float64Type]): ...
class Decimal128Scalar(Scalar[types.Decimal128Type]): ...
class Decimal256Scalar(Scalar[types.Decimal256Type]): ...
class Date32Scalar(Scalar[types.Date32Type]): ...

class Date64Scalar(Scalar[types.Date64Type]):
    @property
    def value(self) -> dt.date | None: ...

class Time32Scalar(Scalar[types.Time32Type[_Time32Unit]]):
    @property
    def value(self) -> dt.time | None: ...

class Time64Scalar(Scalar[types.Time64Type[_Time64Unit]]):
    @property
    def value(self) -> dt.time | None: ...

class TimestampScalar(Scalar[types.TimestampType[_Unit, _Tz]]):
    @property
    def value(self) -> int | None: ...

class DurationScalar(Scalar[types.DurationType[_Unit]]):
    @property
    def value(self) -> dt.timedelta | None: ...

class MonthDayNanoIntervalScalar(Scalar[types.MonthDayNanoIntervalType]):
    @property
    def value(self) -> MonthDayNano | None: ...

class BinaryScalar(Scalar[types.BinaryType]):
    def as_buffer(self) -> Buffer: ...

class LargeBinaryScalar(Scalar[types.LargeBinaryType]):
    def as_buffer(self) -> Buffer: ...

class FixedSizeBinaryScalar(Scalar[types.FixedSizeBinaryType]):
    def as_buffer(self) -> Buffer: ...

class StringScalar(Scalar[types.StringType]):
    def as_buffer(self) -> Buffer: ...

class LargeStringScalar(Scalar[types.LargeStringType]):
    def as_buffer(self) -> Buffer: ...

class BinaryViewScalar(Scalar[types.BinaryViewType]):
    def as_buffer(self) -> Buffer: ...

class StringViewScalar(Scalar[types.StringViewType]):
    def as_buffer(self) -> Buffer: ...

class ListScalar(Scalar[types.ListType[_DataType_CoT]]):
    @property
    def values(self) -> Array | None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i: int) -> Scalar[_DataType_CoT]: ...
    def __iter__(self) -> Iterator[Array]: ...

class FixedSizeListScalar(Scalar[types.FixedSizeListType[_DataType_CoT, types._Size]]):
    @property
    def values(self) -> Array | None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i: int) -> Scalar[_DataType_CoT]: ...
    def __iter__(self) -> Iterator[Array]: ...

class LargeListScalar(Scalar[types.LargeListType[_DataType_CoT]]):
    @property
    def values(self) -> Array | None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i: int) -> Scalar[_DataType_CoT]: ...
    def __iter__(self) -> Iterator[Array]: ...

class ListViewScalar(Scalar[types.ListViewType[_DataType_CoT]]):
    @property
    def values(self) -> Array | None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i: int) -> Scalar[_DataType_CoT]: ...
    def __iter__(self) -> Iterator[Array]: ...

class LargeListViewScalar(Scalar[types.LargeListViewType[_DataType_CoT]]):
    @property
    def values(self) -> Array | None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i: int) -> Scalar[_DataType_CoT]: ...
    def __iter__(self) -> Iterator[Array]: ...

class StructScalar(Scalar[types.StructType], collections.abc.Mapping[str, Scalar]):
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[str]: ...
    def __getitem__(self, __key: str) -> Scalar[Any]: ...  # type: ignore[override]
    def _as_py_tuple(self) -> list[tuple[str, Any]]: ...

class MapScalar(Scalar[types.MapType[types._K, types._ValueT]]):
    @property
    def values(self) -> Array | None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i: int) -> tuple[Scalar[types._K], types._ValueT, Any]: ...
    @overload
    def __iter__(
        self: Scalar[
            types.MapType[types._BasicDataType[_AsPyTypeK], types._BasicDataType[_AsPyTypeV]]
        ],
    ) -> Iterator[tuple[_AsPyTypeK, _AsPyTypeV]]: ...
    @overload
    def __iter__(
        self: Scalar[types.MapType[Any, types._BasicDataType[_AsPyTypeV]],],
    ) -> Iterator[tuple[Any, _AsPyTypeV]]: ...
    @overload
    def __iter__(
        self: Scalar[types.MapType[types._BasicDataType[_AsPyTypeK], Any],],
    ) -> Iterator[tuple[_AsPyTypeK, Any]]: ...

class DictionaryScalar(Scalar[types.DictionaryType[types._IndexT, types._BasicValueT]]):
    @property
    def index(self) -> Scalar[types._IndexT]: ...
    @property
    def value(self) -> Scalar[types._BasicValueT]: ...
    @property
    def dictionary(self) -> Array: ...

class RunEndEncodedScalar(Scalar[types.RunEndEncodedType[types._RunEndType, types._BasicValueT]]):
    @property
    def value(self) -> tuple[int, types._BasicValueT] | None: ...

class UnionScalar(Scalar[types.UnionType]):
    @property
    def value(self) -> Any | None: ...
    @property
    def type_code(self) -> str: ...

class ExtensionScalar(Scalar[types.ExtensionType]):
    @property
    def value(self) -> Any | None: ...
    @staticmethod
    def from_storage(typ: types.BaseExtensionType, value) -> ExtensionScalar: ...

class FixedShapeTensorScalar(ExtensionScalar):
    def to_numpy(self) -> np.ndarray: ...
    def to_tensor(self) -> Tensor: ...

_V = TypeVar("_V")

CollectionValue: TypeAlias = list[_V | None] | tuple[_V | None, ...] | set[_V | None]

@overload
def scalar(
    value: str, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> StringScalar: ...
@overload
def scalar(
    value: bytes, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> BinaryScalar: ...
@overload
def scalar(
    value: bool, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> BooleanScalar: ...
@overload
def scalar(
    value: int, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> Int64Scalar: ...
@overload
def scalar(
    value: float, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> DoubleScalar: ...
@overload
def scalar(
    value: Decimal, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> Decimal128Scalar: ...
@overload
def scalar(
    value: dt.datetime, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> TimestampScalar: ...
@overload
def scalar(
    value: dt.date, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> Date32Scalar: ...
@overload
def scalar(
    value: dt.time, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> Time64Scalar: ...
@overload
def scalar(
    value: dt.timedelta, *, from_pandas: bool | None = None, memory_pool: MemoryPool | None = None
) -> DurationScalar: ...
@overload
def scalar(
    value: MonthDayNano,
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> MonthDayNanoIntervalScalar: ...
@overload
def scalar(
    value: Mapping[str, Any],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> StructScalar: ...
@overload
def scalar(
    value: CollectionValue[str],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.StringType]]: ...
@overload
def scalar(
    value: CollectionValue[bytes],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.BinaryType]]: ...
@overload
def scalar(
    value: CollectionValue[bool],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.BoolType]]: ...
@overload
def scalar(
    value: CollectionValue[int],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.Int64Type]]: ...
@overload
def scalar(
    value: CollectionValue[float],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.Float64Type]]: ...
@overload
def scalar(
    value: CollectionValue[Decimal],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.Decimal128Type]]: ...
@overload
def scalar(
    value: CollectionValue[dt.datetime],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.TimestampType]]: ...
@overload
def scalar(
    value: CollectionValue[dt.date],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.Date32Type]]: ...
@overload
def scalar(
    value: CollectionValue[dt.time],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.Time32Type]]: ...
@overload
def scalar(
    value: CollectionValue[dt.timedelta],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.DurationType]]: ...
@overload
def scalar(
    value: CollectionValue[MonthDayNano],
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[types.ListType[types.MonthDayNanoIntervalType]]: ...
@overload
def scalar(
    value: CollectionValue,
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> ListScalar[Any]: ...
@overload
def scalar(
    value: Any,
    type: _DataTypeT,
    *,
    from_pandas: bool | None = None,
    memory_pool: MemoryPool | None = None,
) -> Scalar[_DataTypeT]: ...

__all__ = [
    "Scalar",
    "_NULL",
    "NA",
    "NullScalar",
    "BooleanScalar",
    "UInt8Scalar",
    "Int8Scalar",
    "UInt16Scalar",
    "Int16Scalar",
    "UInt32Scalar",
    "Int32Scalar",
    "UInt64Scalar",
    "Int64Scalar",
    "HalfFloatScalar",
    "FloatScalar",
    "DoubleScalar",
    "Decimal128Scalar",
    "Decimal256Scalar",
    "Date32Scalar",
    "Date64Scalar",
    "Time32Scalar",
    "Time64Scalar",
    "TimestampScalar",
    "DurationScalar",
    "MonthDayNanoIntervalScalar",
    "BinaryScalar",
    "LargeBinaryScalar",
    "FixedSizeBinaryScalar",
    "StringScalar",
    "LargeStringScalar",
    "BinaryViewScalar",
    "StringViewScalar",
    "ListScalar",
    "FixedSizeListScalar",
    "LargeListScalar",
    "ListViewScalar",
    "LargeListViewScalar",
    "StructScalar",
    "MapScalar",
    "DictionaryScalar",
    "RunEndEncodedScalar",
    "UnionScalar",
    "ExtensionScalar",
    "FixedShapeTensorScalar",
    "scalar",
]



from typing import Literal

from ._fs import FileSystem

class AzureFileSystem(FileSystem):
    def __init__(
        self,
        account_name: str,
        account_key: str | None = None,
        blob_storage_authority: str | None = None,
        dfs_storage_authority: str | None = None,
        blob_storage_schema: Literal["http", "https"] = "https",
        dfs_storage_schema: Literal["http", "https"] = "https",
    ) -> None: ...



import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import (
    IO,
    Any,
    Callable,
    Generic,
    Iterator,
    Literal,
    NamedTuple,
    TypeVar,
    overload,
)

from _typeshed import StrPath

from . import _csv, _json, _parquet, lib
from ._fs import FileSelector, FileSystem, SupportedFileSystem
from ._stubs_typing import Indices, JoinType, Order
from .acero import ExecNodeOptions
from .compute import Expression
from .ipc import IpcWriteOptions, RecordBatchReader

class Dataset(lib._Weakrefable):
    @property
    def partition_expression(self) -> Expression: ...
    def replace_schema(self, schema: lib.Schema) -> None: ...
    def get_fragments(self, filter: Expression | None = None): ...
    def scanner(
        self,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> Scanner: ...
    def to_batches(
        self,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> Iterator[lib.RecordBatch]: ...
    def to_table(
        self,
        columns: list[str] | dict[str, Expression] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> lib.Table: ...
    def take(
        self,
        indices: Indices,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> lib.Table: ...
    def head(
        self,
        num_rows: int,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> lib.Table: ...
    def count_rows(
        self,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> int: ...
    @property
    def schema(self) -> lib.Schema: ...
    def filter(self, expression: Expression) -> Self: ...
    def sort_by(self, sorting: str | list[tuple[str, Order]], **kwargs) -> InMemoryDataset: ...
    def join(
        self,
        right_dataset: Dataset,
        keys: str | list[str],
        right_keys: str | list[str] | None = None,
        join_type: JoinType = "left outer",
        left_suffix: str | None = None,
        right_suffix: str | None = None,
        coalesce_keys: bool = True,
        use_threads: bool = True,
    ) -> InMemoryDataset: ...
    def join_asof(
        self,
        right_dataset: Dataset,
        on: str,
        by: str | list[str],
        tolerance: int,
        right_on: str | list[str] | None = None,
        right_by: str | list[str] | None = None,
    ) -> InMemoryDataset: ...

class InMemoryDataset(Dataset): ...

class UnionDataset(Dataset):
    @property
    def children(self) -> list[Dataset]: ...

class FileSystemDataset(Dataset):
    def __init__(
        self,
        fragments: list[Fragment],
        schema: lib.Schema,
        format: FileFormat,
        filesystem: SupportedFileSystem | None = None,
        root_partition: Expression | None = None,
    ) -> None: ...
    @classmethod
    def from_paths(
        cls,
        paths: list[str],
        schema: lib.Schema | None = None,
        format: FileFormat | None = None,
        filesystem: SupportedFileSystem | None = None,
        partitions: list[Expression] | None = None,
        root_partition: Expression | None = None,
    ) -> FileSystemDataset: ...
    @property
    def filesystem(self) -> FileSystem: ...
    @property
    def partitioning(self) -> Partitioning | None: ...
    @property
    def files(self) -> list[str]: ...
    @property
    def format(self) -> FileFormat: ...

class FileWriteOptions(lib._Weakrefable):
    @property
    def format(self) -> FileFormat: ...

class FileFormat(lib._Weakrefable):
    def inspect(
        self, file: StrPath | IO, filesystem: SupportedFileSystem | None = None
    ) -> lib.Schema: ...
    def make_fragment(
        self,
        file: StrPath | IO,
        filesystem: SupportedFileSystem | None = None,
        partition_expression: Expression | None = None,
        *,
        file_size: int | None = None,
    ) -> Fragment: ...
    def make_write_options(self) -> FileWriteOptions: ...
    @property
    def default_extname(self) -> str: ...
    @property
    def default_fragment_scan_options(self) -> FragmentScanOptions: ...
    @default_fragment_scan_options.setter
    def default_fragment_scan_options(self, options: FragmentScanOptions) -> None: ...

class Fragment(lib._Weakrefable):
    @property
    def physical_schema(self) -> lib.Schema: ...
    @property
    def partition_expression(self) -> Expression: ...
    def scanner(
        self,
        schema: lib.Schema | None = None,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> Scanner: ...
    def to_batches(
        self,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> Iterator[lib.RecordBatch]: ...
    def to_table(
        self,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> lib.Table: ...
    def take(
        self,
        indices: Indices,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> lib.Table: ...
    def head(
        self,
        num_rows: int,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> lib.Table: ...
    def count_rows(
        self,
        columns: list[str] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> int: ...

class FileFragment(Fragment):
    def open(self) -> lib.NativeFile: ...
    @property
    def path(self) -> str: ...
    @property
    def filesystem(self) -> FileSystem: ...
    @property
    def buffer(self) -> lib.Buffer: ...
    @property
    def format(self) -> FileFormat: ...

class FragmentScanOptions(lib._Weakrefable):
    @property
    def type_name(self) -> str: ...

class IpcFileWriteOptions(FileWriteOptions):
    @property
    def write_options(self) -> IpcWriteOptions: ...
    @write_options.setter
    def write_options(self, write_options: IpcWriteOptions) -> None: ...

class IpcFileFormat(FileFormat):
    def equals(self, other: IpcFileFormat) -> bool: ...
    def make_write_options(self, **kwargs) -> IpcFileWriteOptions: ...
    @property
    def default_extname(self) -> str: ...

class FeatherFileFormat(IpcFileFormat): ...

class CsvFileFormat(FileFormat):
    def __init__(
        self,
        parse_options: _csv.ParseOptions | None = None,
        default_fragment_scan_options: CsvFragmentScanOptions | None = None,
        convert_options: _csv.ConvertOptions | None = None,
        read_options: _csv.ReadOptions | None = None,
    ) -> None: ...
    def make_write_options(self) -> _csv.WriteOptions: ...  # type: ignore[override]
    @property
    def parse_options(self) -> _csv.ParseOptions: ...
    @parse_options.setter
    def parse_options(self, parse_options: _csv.ParseOptions) -> None: ...
    def equals(self, other: CsvFileFormat) -> bool: ...

class CsvFragmentScanOptions(FragmentScanOptions):
    convert_options: _csv.ConvertOptions
    read_options: _csv.ReadOptions

    def __init__(
        self, convert_options: _csv.ConvertOptions, read_options: _csv.ReadOptions
    ) -> None: ...
    def equals(self, other: CsvFragmentScanOptions) -> bool: ...

class CsvFileWriteOptions(FileWriteOptions):
    write_options: _csv.WriteOptions

class JsonFileFormat(FileFormat):
    def __init__(
        self,
        default_fragment_scan_options: JsonFragmentScanOptions | None = None,
        parse_options: _json.ParseOptions | None = None,
        read_options: _json.ReadOptions | None = None,
    ) -> None: ...
    def equals(self, other: JsonFileFormat) -> bool: ...

class JsonFragmentScanOptions(FragmentScanOptions):
    parse_options: _json.ParseOptions
    read_options: _json.ReadOptions
    def __init__(
        self, parse_options: _json.ParseOptions, read_options: _json.ReadOptions
    ) -> None: ...
    def equals(self, other: JsonFragmentScanOptions) -> bool: ...

class Partitioning(lib._Weakrefable):
    def parse(self, path: str) -> Expression: ...
    @property
    def schema(self) -> lib.Schema: ...

class PartitioningFactory(lib._Weakrefable):
    @property
    def type_name(self) -> str: ...

class KeyValuePartitioning(Partitioning):
    @property
    def dictionaries(self) -> list[lib.Array | None]: ...

class DirectoryPartitioning(KeyValuePartitioning):
    @staticmethod
    def discover(
        field_names: list[str] | None = None,
        infer_dictionary: bool = False,
        max_partition_dictionary_size: int = 0,
        schema: lib.Schema | None = None,
        segment_encoding: Literal["uri", "none"] = "uri",
    ) -> PartitioningFactory: ...
    def __init__(
        self,
        schema: lib.Schema,
        dictionaries: dict[str, lib.Array] | None = None,
        segment_encoding: Literal["uri", "none"] = "uri",
    ) -> None: ...

class HivePartitioning(KeyValuePartitioning):
    def __init__(
        self,
        schema: lib.Schema,
        dictionaries: dict[str, lib.Array] | None = None,
        null_fallback: str = "__HIVE_DEFAULT_PARTITION__",
        segment_encoding: Literal["uri", "none"] = "uri",
    ) -> None: ...
    @staticmethod
    def discover(
        infer_dictionary: bool = False,
        max_partition_dictionary_size: int = 0,
        null_fallback="__HIVE_DEFAULT_PARTITION__",
        schema: lib.Schema | None = None,
        segment_encoding: Literal["uri", "none"] = "uri",
    ) -> PartitioningFactory: ...

class FilenamePartitioning(KeyValuePartitioning):
    def __init__(
        self,
        schema: lib.Schema,
        dictionaries: dict[str, lib.Array] | None = None,
        segment_encoding: Literal["uri", "none"] = "uri",
    ) -> None: ...
    @staticmethod
    def discover(
        field_names: list[str] | None = None,
        infer_dictionary: bool = False,
        schema: lib.Schema | None = None,
        segment_encoding: Literal["uri", "none"] = "uri",
    ) -> PartitioningFactory: ...

class DatasetFactory(lib._Weakrefable):
    root_partition: Expression
    def finish(self, schema: lib.Schema | None = None) -> Dataset: ...
    def inspect(self) -> lib.Schema: ...
    def inspect_schemas(self) -> list[lib.Schema]: ...

class FileSystemFactoryOptions(lib._Weakrefable):
    partitioning: Partitioning
    partitioning_factory: PartitioningFactory
    partition_base_dir: str
    exclude_invalid_files: bool
    selector_ignore_prefixes: list[str]

    def __init__(
        self,
        artition_base_dir: str | None = None,
        partitioning: Partitioning | PartitioningFactory | None = None,
        exclude_invalid_files: bool = True,
        selector_ignore_prefixes: list[str] | None = None,
    ) -> None: ...

class FileSystemDatasetFactory(DatasetFactory):
    def __init__(
        self,
        filesystem: SupportedFileSystem,
        paths_or_selector: FileSelector,
        format: FileFormat,
        options: FileSystemFactoryOptions | None = None,
    ) -> None: ...

class UnionDatasetFactory(DatasetFactory):
    def __init__(self, factories: list[DatasetFactory]) -> None: ...

_RecordBatchT = TypeVar("_RecordBatchT", bound=lib.RecordBatch)

class RecordBatchIterator(lib._Weakrefable, Generic[_RecordBatchT]):
    def __iter__(self) -> Self: ...
    def __next__(self) -> _RecordBatchT: ...

class TaggedRecordBatch(NamedTuple):
    record_batch: lib.RecordBatch
    fragment: Fragment

class TaggedRecordBatchIterator(lib._Weakrefable):
    def __iter__(self) -> Self: ...
    def __next__(self) -> TaggedRecordBatch: ...

class Scanner(lib._Weakrefable):
    @staticmethod
    def from_dataset(
        dataset: Dataset,
        *,
        columns: list[str] | dict[str, Expression] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> Scanner: ...
    @staticmethod
    def from_fragment(
        fragment: Fragment,
        *,
        schema: lib.Schema | None = None,
        columns: list[str] | dict[str, Expression] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> Scanner: ...
    @overload
    @staticmethod
    def from_batches(
        source: Iterator[lib.RecordBatch],
        *,
        schema: lib.Schema,
        columns: list[str] | dict[str, Expression] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> Scanner: ...
    @overload
    @staticmethod
    def from_batches(
        source: RecordBatchReader,
        *,
        columns: list[str] | dict[str, Expression] | None = None,
        filter: Expression | None = None,
        batch_size: int = ...,
        batch_readahead: int = 16,
        fragment_readahead: int = 4,
        fragment_scan_options: FragmentScanOptions | None = None,
        use_threads: bool = True,
        memory_pool: lib.MemoryPool | None = None,
    ) -> Scanner: ...
    @property
    def dataset_schema(self) -> lib.Schema: ...
    @property
    def projected_schema(self) -> lib.Schema: ...
    def to_batches(self) -> Iterator[lib.RecordBatch]: ...
    def scan_batches(self) -> TaggedRecordBatchIterator: ...
    def to_table(self) -> lib.Table: ...
    def take(self, indices: Indices) -> lib.Table: ...
    def head(self, num_rows: int) -> lib.Table: ...
    def count_rows(self) -> int: ...
    def to_reader(self) -> RecordBatchReader: ...

def get_partition_keys(partition_expression: Expression) -> dict[str, Any]: ...

class WrittenFile(lib._Weakrefable):
    def __init__(self, path: str, metadata: _parquet.FileMetaData | None, size: int) -> None: ...

def _filesystemdataset_write(
    data: Scanner,
    base_dir: StrPath,
    basename_template: str,
    filesystem: SupportedFileSystem,
    partitioning: Partitioning,
    file_options: FileWriteOptions,
    max_partitions: int,
    file_visitor: Callable[[str], None],
    existing_data_behavior: Literal["error", "overwrite_or_ignore", "delete_matching"],
    max_open_files: int,
    max_rows_per_file: int,
    min_rows_per_group: int,
    max_rows_per_group: int,
    create_dir: bool,
): ...

class _ScanNodeOptions(ExecNodeOptions):
    def _set_options(self, dataset: Dataset, scan_options: dict) -> None: ...

class ScanNodeOptions(_ScanNodeOptions):
    def __init__(self, dataset: Dataset, **kwargs) -> None: ...



from typing import Iterable

from pyarrow.lib import MemoryPool, _Weakrefable

from .array import StringArray, StringViewArray

class StringBuilder(_Weakrefable):
    def __init__(self, memory_pool: MemoryPool | None = None) -> None: ...
    def append(self, value: str | bytes | None): ...
    def append_values(self, values: Iterable[str | bytes | None]): ...
    def finish(self) -> StringArray: ...
    @property
    def null_count(self) -> int: ...
    def __len__(self) -> int: ...

class StringViewBuilder(_Weakrefable):
    def __init__(self, memory_pool: MemoryPool | None = None) -> None: ...
    def append(self, value: str | bytes | None): ...
    def append_values(self, values: Iterable[str | bytes | None]): ...
    def finish(self) -> StringViewArray: ...
    @property
    def null_count(self) -> int: ...
    def __len__(self) -> int: ...

__all__ = ["StringBuilder", "StringViewBuilder"]



import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import IO, Literal

from _typeshed import StrPath

from . import _orc
from ._fs import SupportedFileSystem
from .lib import KeyValueMetadata, NativeFile, RecordBatch, Schema, Table

class ORCFile:
    reader: _orc.ORCReader
    def __init__(self, source: StrPath | NativeFile | IO) -> None: ...
    @property
    def metadata(self) -> KeyValueMetadata: ...
    @property
    def schema(self) -> Schema: ...
    @property
    def nrows(self) -> int: ...
    @property
    def nstripes(self) -> int: ...
    @property
    def file_version(self) -> str: ...
    @property
    def software_version(self) -> str: ...
    @property
    def compression(self) -> Literal["UNCOMPRESSED", "ZLIB", "SNAPPY", "LZ4", "ZSTD"]: ...
    @property
    def compression_size(self) -> int: ...
    @property
    def writer(self) -> str: ...
    @property
    def writer_version(self) -> str: ...
    @property
    def row_index_stride(self) -> int: ...
    @property
    def nstripe_statistics(self) -> int: ...
    @property
    def content_length(self) -> int: ...
    @property
    def stripe_statistics_length(self) -> int: ...
    @property
    def file_footer_length(self) -> int: ...
    @property
    def file_postscript_length(self) -> int: ...
    @property
    def file_length(self) -> int: ...
    def read_stripe(self, n: int, columns: list[str] | None = None) -> RecordBatch: ...
    def read(self, columns: list[str] | None = None) -> Table: ...

class ORCWriter:
    writer: _orc.ORCWriter
    is_open: bool
    def __init__(
        self,
        where: StrPath | NativeFile | IO,
        *,
        file_version: str = "0.12",
        batch_size: int = 1024,
        stripe_size: int = 64 * 1024 * 1024,
        compression: Literal["UNCOMPRESSED", "ZLIB", "SNAPPY", "LZ4", "ZSTD"] = "UNCOMPRESSED",
        compression_block_size: int = 65536,
        compression_strategy: Literal["COMPRESSION", "SPEED"] = "SPEED",
        row_index_stride: int = 10000,
        padding_tolerance: float = 0.0,
        dictionary_key_size_threshold: float = 0.0,
        bloom_filter_columns: list[int] | None = None,
        bloom_filter_fpp: float = 0.05,
    ): ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *args, **kwargs) -> None: ...
    def write(self, table: Table) -> None: ...
    def close(self) -> None: ...

def read_table(
    source: StrPath | NativeFile | IO,
    columns: list[str] | None = None,
    filesystem: SupportedFileSystem | None = None,
) -> Table: ...
def write_table(
    table: Table,
    where: StrPath | NativeFile | IO,
    *,
    file_version: str = "0.12",
    batch_size: int = 1024,
    stripe_size: int = 64 * 1024 * 1024,
    compression: Literal["UNCOMPRESSED", "ZLIB", "SNAPPY", "LZ4", "ZSTD"] = "UNCOMPRESSED",
    compression_block_size: int = 65536,
    compression_strategy: Literal["COMPRESSION", "SPEED"] = "SPEED",
    row_index_stride: int = 10000,
    padding_tolerance: float = 0.0,
    dictionary_key_size_threshold: float = 0.0,
    bloom_filter_columns: list[int] | None = None,
    bloom_filter_fpp: float = 0.05,
) -> None: ...
